import gzip
import os
import html
import ftfy
import re
import string
import tensorflow as tf
import tensorflow_text as tf_text
from keras.src.utils import data_utils
from tensorflow_hub import KerasLayer

DEFAULT_CONTEXT_LENGTH = 77  # Default context length for OpenAI CLIP


def basic_clean(text):
    text = ftfy.fix_text(text)
    text = html.unescape(html.unescape(text))
    text = text.strip()

    return text


class SimpleTokenizer:
    def __init__(self, context_length=DEFAULT_CONTEXT_LENGTH):
        self.context_length = context_length

        byte_list, unicode_list = self._bytes_to_unicode()

        vocab_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'bpe_simple_vocab_16e6.txt.gz')
        with gzip.open(vocab_path, 'rt') as vocab_file:
            merges = vocab_file.read().split('\n')
        merges = merges[1:49152 - 256 - 2 + 1]
        self.merge_size = len(merges)
        self.max_len = max([len(m.replace(' ', '').replace('</w>', '').encode('utf-8')) for m in merges])

        vocabulary = unicode_list + [f'{u}</w>' for u in unicode_list]
        vocabulary.extend([m.replace(' ', '') for m in merges])
        self.vocab_size = len(vocabulary)

        # Byte -> unicode mapping
        self.byte2unicode = tf.lookup.StaticHashTable(
            tf.lookup.KeyValueTensorInitializer(byte_list, unicode_list), '')

        # Token -> id mapping
        self.token2id = tf.lookup.StaticHashTable(
            tf.lookup.KeyValueTensorInitializer(vocabulary, range(len(vocabulary))), -1)

        # Pair -> frequency mapping
        self.merge2rank = tf.lookup.StaticHashTable(
            tf.lookup.KeyValueTensorInitializer(merges, range(len(merges))), self.merge_size + 1)

        # Word -> id cache
        # `tf.lookup.experimental.MutableHashTable` does not support string to string mapping.
        # So we first convert to string to an integer key, and use the integer key to find the value.
        self.cache_factors = tf.pow(tf.constant(256, dtype='int64'), tf.range(0, 8, dtype='int64'))[:, None]
        self.cache_id2value = tf.lookup.experimental.MutableHashTable('int64', 'string', '')

    def _bytes_to_unicode(self):
        bs = list(range(ord("!"), ord("~") + 1)) + \
             list(range(ord("¡"), ord("¬") + 1)) + \
             list(range(ord("®"), ord("ÿ") + 1))
        cs = bs[:]
        n = 0
        for b in range(2 ** 8):
            if b not in bs:
                bs.append(b)
                cs.append(2 ** 8 + n)
                n += 1
        cs = [chr(n) for n in cs]
        bs = [n.to_bytes(1, "little") for n in bs]

        return bs, cs

    def __call__(self, inputs):
        with tf.device('cpu'):
            inputs = tf.convert_to_tensor(inputs, 'string')
            if 1 != inputs.shape.rank:
                raise ValueError(f'Expecting a vector of strings. Got shape {inputs.shape}')

            # Lowercase input sentences and separate words with space
            inputs = tf.strings.lower(inputs, encoding='utf-8')
            inputs = tf.strings.regex_replace(
                inputs, '[\v\u000b\u0085\u00a0\u0345\u1680\u2000-\u200a\u2028\u2029\u202f\u205f\u3000]', ' ')
            inputs = tf.strings.regex_replace(
                inputs, r"'s|'t|'re|'ve|'m|'ll|'d|\p{L}+|\p{N}|[^\s\p{L}\p{N}]+", r" \0 ")

            # Split sentences to words
            raw_tokens = tf.strings.split(inputs)

            # Slice long sentences to speedup
            raw_tokens = raw_tokens[..., :self.context_length - 2]

            row_splits = raw_tokens.row_splits
            flat_tokens = raw_tokens.flat_values

            # Get word indices from cache
            cache_lookup = self._cache_lookup(flat_tokens)
            cache_miss = tf.equal(cache_lookup, '') & tf.not_equal(flat_tokens, '')

            # Tokenize unseen words
            tokenized_words = tf.cond(
                tf.math.reduce_any(cache_miss),
                lambda: self._process_unseen(flat_tokens, cache_miss),
                lambda: cache_lookup)

            # Lookup token indices
            tokens = tf.strings.split(tokenized_words)
            tokens = self.token2id.lookup(tokens)

            # Unflatten to match input
            tokens = tf.RaggedTensor.from_row_splits(tokens.flat_values, tf.gather(tokens.row_splits, row_splits))

            # Slice to required sequence length
            if self.context_length:
                tokens = tokens[:, :self.context_length - 2]

            # Add begin-of-sentence and begin-of-sentence tokens
            bos = tf.fill([tokens.nrows(), 1], self.vocab_size)
            eos = tf.fill([tokens.nrows(), 1], self.vocab_size + 1)
            tokens = tf.concat([bos, tokens, eos], axis=1)

            # Make indices dense
            if self.context_length:
                output_shape = tokens.shape.as_list()
                output_shape[-1] = self.context_length
                tokens = tokens.to_tensor(shape=output_shape)

            return tokens

    def _cache_key(self, keys):
        # `tf.fingerprint` converts token to an array of uint8 of length 8, we need to convert it to int64.
        return tf.squeeze(tf.matmul(tf.cast(tf.fingerprint(keys), dtype='int64'), self.cache_factors), -1)

    def _cache_lookup(self, keys):
        ids = self._cache_key(keys)
        result = self.cache_id2value.lookup(ids)
        result.set_shape([None])  # Ensure output shape for graph mode.

        return result

    def _cache_insert(self, keys, values):
        self.cache_id2value.insert(self._cache_key(keys), values)

    def _process_unseen(self, tokens, miss_mask):
        # Process unseen tokens and add to cache
        unseen_tokens = tf.boolean_mask(tokens, miss_mask)
        seen_words = self._bpe_encode(unseen_tokens)
        seen_words = tf.strings.reduce_join(seen_words, axis=1, separator=' ')
        self._cache_insert(unseen_tokens, seen_words)

        return self._cache_lookup(tokens)

    def _bpe_encode(self, words):
        # Split to bytes
        chars = tf.strings.bytes_split(words)

        # Slice long words to speedup
        chars = chars[:, :self.max_len * 2]  # double max length due end-of-word marker changes merge priority

        # Map bytes to unicode
        chars = self.byte2unicode.lookup(chars)

        # Add end-of-word marker
        chars = tf.concat([
            chars[:, :-1], tf.strings.join([chars[:, -1:], '</w>'])], axis=-1)

        # Merge characters
        tokens, _ = tf.while_loop(
            lambda _, mask: tf.math.reduce_any(mask),
            self._bpe_step,
            loop_vars=(chars, tf.ones_like(words, 'bool')),
            shape_invariants=(tf.TensorShape([None, None]), tf.TensorShape([None])))

        return tokens

    @tf.function
    def _bpe_step(self, words, merge_mask):
        merge_mask &= words.row_lengths() > 1

        return tf.cond(
            tf.reduce_any(merge_mask),
            lambda: self._bpe_merge(words, merge_mask),
            lambda: (words, merge_mask))

    def _bpe_merge(self, words, merge_mask):
        # Sub-word pairs for words that possible can be merged
        curr_words = tf.ragged.boolean_mask(words, merge_mask)
        pairs_left, pairs_right = curr_words[:, :-1], curr_words[:, 1:]
        curr_pairs = tf.strings.join([pairs_left, pairs_right], separator=' ')

        # Byte pair ranking in merge rules
        curr_ranks = self.merge2rank.lookup(curr_pairs)
        min_ranks = tf.reduce_min(curr_ranks, axis=-1)
        curr_found = min_ranks < self.merge_size + 1

        # Tokens that cannot be further merged are marked as finished
        found_indices = tf.boolean_mask(tf.range(tf.size(merge_mask)), merge_mask)
        merge_mask = tf.tensor_scatter_nd_update(merge_mask, found_indices[:, None], curr_found)

        return tf.cond(
            tf.reduce_any(merge_mask),
            lambda: self._bpe_replace(words, merge_mask, curr_ranks, min_ranks, curr_found),
            lambda: (words, merge_mask))

    def _bpe_replace(self, words, merge_mask, pair_ranks, min_ranks, found_mask):
        # Mask for right sub-word in pairs that definitely should be merged
        merge_left = tf.equal(pair_ranks, min_ranks[:, None])
        merge_left = tf.ragged.boolean_mask(merge_left, found_mask)
        merge_shift = tf.zeros([merge_left.nrows(), 1], 'bool')
        merge_right = tf.concat([merge_shift, merge_left], axis=1)

        # All sub-words indices
        words_size = tf.size(words.flat_values)
        words_range = tf.range(words_size)
        word_idx = words.with_flat_values(words_range)

        # Separate `skip` and `merge` sub-word indices
        curr_idx = tf.ragged.boolean_mask(word_idx, merge_mask)
        skip_idx = tf.ragged.boolean_mask(word_idx, ~merge_mask)

        # Segment shift from sub-word merge mask
        seg_shift = words.with_flat_values(tf.dynamic_stitch(
            [curr_idx.flat_values, skip_idx.flat_values],
            [tf.cast(merge_right.flat_values, words_size.dtype), tf.zeros_like(skip_idx.flat_values)]))

        # Merge sub-word pairs
        segment_ids = words_range - tf.cumsum(seg_shift.flat_values)
        num_segments = tf.reduce_max(segment_ids) + 1
        merged_words = tf.strings.unsorted_segment_join(words.flat_values, segment_ids, num_segments, '')

        # Merged sub-words
        words_lengths = words.row_lengths()
        merged_lengths = words_lengths - tf.cast(tf.reduce_sum(seg_shift, axis=-1), words_lengths.dtype)
        words = tf.RaggedTensor.from_row_lengths(merged_words, merged_lengths)

        return words, merge_mask


class SentencePieceTokenizer:
    def __init__(self, tokenizer_name, context_length=DEFAULT_CONTEXT_LENGTH, clean=None):
        model_url = f'https://github.com/shkarupa-alex/tfclip/releases/download/{tokenizer_name}.model'
        model_hash = model_url.split('__')[-1].replace('.model', '')
        model_path = data_utils.get_file(origin=model_url, file_hash=model_hash, cache_subdir='tfclip')

        with tf.io.gfile.GFile(model_path, 'rb') as f:
            model = f.read()

        self.tokenizer = tf_text.SentencepieceTokenizer(model)
        self.context_length = context_length

        self.canonicalize = None
        if 'canonicalize' == clean:
            punctuation = re.escape(string.punctuation.replace(',', ''))
            self.canonicalize = f'[{punctuation}]+'

    def __call__(self, inputs):
        with tf.device('cpu'):
            inputs = tf.convert_to_tensor(inputs, 'string')
            if 1 != inputs.shape.rank:
                raise ValueError(f'Expecting a vector of strings. Got shape {inputs.shape}')

            # https://github.com/google-research/big_vision/issues/79
            inputs = tf.strings.lower(inputs)  # , encoding='utf-8'

            if self.canonicalize:
                inputs = tf.strings.regex_replace(inputs, self.canonicalize, '')

            tokens = self.tokenizer.tokenize(inputs)

            if self.context_length:
                output_shape = tokens.shape.as_list()
                output_shape[-1] = self.context_length - 1
                tokens = tokens[:, :self.context_length - 1]
                tokens = tokens.to_tensor(default_value=1, shape=output_shape)
                tokens.set_shape(output_shape)
                tokens = tf.pad(tokens, [[0, 0], [0, 1]], constant_values=1)

            return tokens


class TensorflowHubTokenizer:
    def __init__(self, tokenizer_name, context_length=DEFAULT_CONTEXT_LENGTH, strip_sep_token=False):
        self.tokenizer = KerasLayer(f'https://tfhub.dev/tensorflow/{tokenizer_name}')
        self.context_length = context_length
        self.strip_sep_token = strip_sep_token

    def __call__(self, inputs):
        with tf.device('cpu'):
            inputs = tf.convert_to_tensor(inputs, 'string')
            if 1 != inputs.shape.rank:
                raise ValueError(f'Expecting a vector of strings. Got shape {inputs.shape}')

            tokens = self.tokenizer(inputs)['input_word_ids']

            if self.strip_sep_token:
                tokens = tf.where(tf.equal(tokens, int(self.strip_sep_token)), 0, tokens)

            if self.context_length:
                output_shape = tokens.shape.as_list()
                output_shape[-1] = self.context_length - 1
                tokens = tokens[:, :self.context_length - 1]
                tokens.set_shape(output_shape)
                tokens = tf.pad(tokens, [[0, 0], [0, 1]])

            return tokens



import tensorflow as tf
from keras import layers
from keras.saving import register_keras_serializable
from keras.src.utils.tf_utils import shape_type_conversion


@register_keras_serializable(package='TFCLIP')
class TextGlobalPool(layers.Layer):
    def __init__(self, mode, **kwargs):
        super().__init__(**kwargs)
        self.input_spec = [layers.InputSpec(ndim=3), layers.InputSpec(ndim=2, dtype='int64')]

        if mode not in {'first', 'last', 'argmax', 'none'}:
            raise ValueError(f'Unsupported pooling mode: {mode}')

        self.mode = mode

    def call(self, inputs, *args, **kwargs):
        tokens, texts = inputs

        if 'first' == self.mode:
            pooled, tokens = tokens[:, 0], tokens[:, 1:]
        elif 'last' == self.mode:
            pooled, tokens = tokens[:, -1], tokens[:, :-1]
        elif 'argmax' == self.mode:
            lendiff = tf.shape(tokens)[1] - tf.shape(texts)[1]
            indices = tf.cast(texts > 0, 'int32')
            indices = tf.reduce_sum(indices, axis=1) + lendiff - 1
            pooled = tf.gather(tokens, indices, batch_dims=1)
            tokens = tokens
        else:
            pooled = tokens = tokens

        return pooled, tokens

    @shape_type_conversion
    def compute_output_shape(self, input_shape):
        if 'none' == self.mode:
            return input_shape[0], input_shape[0]

        pooled_shape = input_shape[0][:1] + input_shape[0][2:]
        if 'argmax' == self.mode:
            return pooled_shape, input_shape[0]

        length = None if input_shape[0][1] is None else input_shape[0][1] - 1
        tokens_shape = input_shape[0][:1] + (length,) + input_shape[0][2:]

        return pooled_shape, tokens_shape

    def get_config(self):
        config = super().get_config()
        config.update({'mode': self.mode})

        return config



from tfclip.factory import list_models, create_model, create_model_and_transforms, get_tokenizer
from tfclip.pretrain import list_pretrained, list_pretrained_models_by_tag, list_pretrained_tags_by_model



import json
import re
from copy import deepcopy
from dataclasses import asdict
from keras.src.utils import data_utils
from pathlib import Path
from tfclip.tokenizer import DEFAULT_CONTEXT_LENGTH, SimpleTokenizer, SentencePieceTokenizer, TensorflowHubTokenizer
from tfclip.transform import PreprocessCfg, merge_preprocess_dict, image_transform
from tfclip.model import CLIP
from tfclip.pretrain import get_pretrained_cfg, list_pretrained_tags_by_model, get_pretrained_url

HF_HUB_PREFIX = 'hf-hub:'
_MODEL_CONFIG_PATHS = [Path(__file__).parent / f'configs/']
_MODEL_CONFIGS = {}


def _natural_key(string_):
    return [int(s) if s.isdigit() else s for s in re.split(r'(\d+)', string_.lower())]


def _rescan_model_configs():
    global _MODEL_CONFIGS

    config_ext = ('.json',)
    config_files = []
    for config_path in _MODEL_CONFIG_PATHS:
        if config_path.is_file() and config_path.suffix in config_ext:
            config_files.append(config_path)
        elif config_path.is_dir():
            for ext in config_ext:
                config_files.extend(config_path.glob(f'*{ext}'))

    for cf in config_files:
        with open(cf, 'r') as f:
            model_cfg = json.load(f)
            disabled = model_cfg.pop('disabled', False)

            if not disabled and all(a in model_cfg for a in ('embed_dim', 'vision_cfg', 'text_cfg')):
                _MODEL_CONFIGS[cf.stem] = model_cfg

    _MODEL_CONFIGS = {k: v for k, v in sorted(_MODEL_CONFIGS.items(), key=lambda x: _natural_key(x[0]))}


_rescan_model_configs()  # initial populate of model config registry


def list_models():
    return list(_MODEL_CONFIGS.keys())


def get_model_config(model_name):
    if model_name in _MODEL_CONFIGS:
        return deepcopy(_MODEL_CONFIGS[model_name])
    else:
        return None


def get_tokenizer(model_name='', context_length=None, **kwargs):
    if model_name.startswith(HF_HUB_PREFIX):
        raise ValueError(f'Unsupported text tokenizer: {model_name}')
    else:
        config = get_model_config(model_name)
        if config is None:
            raise ValueError(f'No valid model config found for {model_name}')

    text_config = config.get('text_cfg', {})
    tokenizer_kwargs = dict(text_config.get('tokenizer_kwargs', {}), **kwargs)

    if context_length is None:
        context_length = text_config.get('context_length', DEFAULT_CONTEXT_LENGTH)
    tokenizer_kwargs['context_length'] = context_length

    if 'hf_tokenizer_name' in text_config:
        raise ValueError(f'Unsupported text tokenizer: {text_config["hf_tokenizer_name"]}')
    elif 'sp_tokenizer_name' in text_config:
        tokenizer = SentencePieceTokenizer(text_config['sp_tokenizer_name'], **tokenizer_kwargs)
    elif 'hub_tokenizer_name' in text_config:
        tokenizer = TensorflowHubTokenizer(text_config['hub_tokenizer_name'], **tokenizer_kwargs)
    else:
        tokenizer = SimpleTokenizer(**tokenizer_kwargs)

    return tokenizer


def create_model(model_name, pretrained=None, weights_path=None, **model_kwargs):
    if model_name.startswith(HF_HUB_PREFIX):
        raise ValueError(f'Unsupported model: {model_name}')

    model_name = model_name.replace('/', '-')

    model_cfg = get_model_config(model_name)
    if model_cfg is None:
        raise ValueError(f'Model config for {model_name} not found')

    if 'timm_model_name' in model_cfg.get('vision_cfg', {}):
        raise ValueError(f'Unsupported image encoder: {model_cfg["vision_cfg"]["timm_model_name"]}')

    if 'hf_model_name' in model_cfg.get('text_cfg', {}):
        raise ValueError(f'Unsupported text encoder: {model_cfg["text_cfg"]["hf_model_name"]}')

    allowed_weights = list_pretrained_tags_by_model(model_name)
    if pretrained is not None and pretrained not in allowed_weights:
        raise ValueError(
            f'Required combination of model and weights is not available. '
            f'Available weights for {model_name} are: {allowed_weights}')

    preprocess_cfg = asdict(PreprocessCfg())
    if pretrained is not None:
        pretrained_cfg = get_pretrained_cfg(model_name, pretrained)
        preprocess_cfg = merge_preprocess_dict(preprocess_cfg, pretrained_cfg)

    bias_init = model_cfg.pop('init_logit_bias', None)
    custom_text = model_cfg.pop('custom_text', False)

    model_cfg = dict(model_cfg, **model_kwargs)
    model_cfg.pop('multimodal_cfg', None)

    model = CLIP(
        **model_cfg, img_mean=preprocess_cfg['mean'], img_std=preprocess_cfg['std'], bias_init=bias_init,
        custom_text=custom_text)

    if pretrained is not None:
        if weights_path is None:
            weights_url = get_pretrained_url(model_name, pretrained)
            if weights_url is None:
                raise ValueError(
                    f'Pretrained weighs "{pretrained}" for model "{model_name}" are not ready and can\'t be '
                    f'downloaded. You can convert them locally with `convert_weights.py` script (open_clip should be '
                    f'installed) and supply with `weights_path` argument.')

            weights_hash = weights_url.split('__')[-1].replace('.h5', '')
            weights_path = data_utils.get_file(origin=weights_url, file_hash=weights_hash, cache_subdir='tfclip')
        model.load_weights(weights_path)

    return model


def create_model_and_transforms(model_name, pretrained=None, weights_path=None, **model_kwargs):
    model = create_model(model_name, pretrained=pretrained, weights_path=weights_path, **model_kwargs)

    pretrained_cfg = {} if pretrained is None else get_pretrained_cfg(model_name, pretrained)
    pretrained_cfg.pop('version', None)
    pretrained_cfg.pop('sha256', None)

    preprocess_cfg = PreprocessCfg(**pretrained_cfg)
    image_prep = image_transform(preprocess_cfg.size, preprocess_cfg.interpolation, preprocess_cfg.resize_mode)

    text_prep = get_tokenizer(model_name)

    return model, image_prep, text_prep



import numpy as np
from keras import backend, layers, models
from keras.src.applications import imagenet_utils
from tfclip.abspos import ImagePositionEmbedding
from tfclip.attnpool import AttentionalPooler
from tfclip.clstok import AddClassToken, SplitClassToken
from tfclip.lscale import LayerScale
from tfclip.mapool import MultiheadAttentionPooling
from tfclip.qgelu import q_gelu


def VisionTransformer(embed_dim, vision_cfg, quick_gelu, img_mean, img_std, name='vision'):
    num_heads = vision_cfg.width // vision_cfg.head_width

    ln_epsilon = 1.001e-5
    if vision_cfg.norm_kwargs:
        norm_kwargs_ = set(vision_cfg.norm_kwargs.keys()) - {'eps'}
        if set(vision_cfg.norm_kwargs.keys()) - {'eps'}:
            raise ValueError(f'Unsupported normalization arguments in config: {norm_kwargs_}')

        ln_epsilon = vision_cfg.norm_kwargs.get('eps', ln_epsilon)

    if vision_cfg.act_kwargs is not None:
        act_kwargs_ = set(vision_cfg.act_kwargs.keys())
        raise ValueError(f'Unsupported activation arguments in config: {act_kwargs_}')

    if vision_cfg.patch_dropout:
        raise ValueError(f'Unsupported patch dropout in config: {vision_cfg.patch_dropout}')

    if not isinstance(vision_cfg.attentional_pool, bool):
        raise ValueError(f'Unsupported attentional pooling mode in config: {vision_cfg.attentional_pool}')

    if 'learnable' != vision_cfg.pos_embed_type:
        raise ValueError(f'Unsupported positional embedding type in config: {vision_cfg.pos_embed_type}')

    mlp_act = q_gelu if quick_gelu else 'gelu'

    # Define model inputs
    input_shape = imagenet_utils.obtain_input_shape(
        None, default_size=vision_cfg.image_size, min_size=vision_cfg.patch_size, data_format='channel_last',
        require_flatten=False)
    image = layers.Input(shape=input_shape, dtype='uint8', name=f'{name}/images')

    # Define model pipeline
    x = image

    x = layers.Normalization(
        mean=np.array(img_mean) * 255., variance=(np.array(img_std) * 255.) ** 2, name=f'{name}/image/norm')(x)
    x = layers.Conv2D(
        vision_cfg.width, vision_cfg.patch_size, strides=vision_cfg.patch_size, use_bias=vision_cfg.patch_bias,
        name=f'{name}/patch/embed')(x)
    x = layers.Reshape(
        [(vision_cfg.image_size // vision_cfg.patch_size) ** 2, vision_cfg.width], name=f'{name}/patch/flatten')(x)
    if vision_cfg.embed_cls:
        x = AddClassToken(name=f'{name}/patch/cls')(x)
    x = ImagePositionEmbedding(
        vision_cfg.patch_size, vision_cfg.image_size, cls_tok=vision_cfg.embed_cls, name=f'{name}/patch/pos')(x)
    if not vision_cfg.no_ln_pre:
        x = layers.LayerNormalization(epsilon=ln_epsilon, name=f'{name}/patch/norm')(x)

    for i in range(vision_cfg.layers):
        y = layers.LayerNormalization(epsilon=ln_epsilon, name=f'{name}/layer_{i}/attn/norm')(x)
        y = layers.MultiHeadAttention(
            num_heads, vision_cfg.width // num_heads, name=f'{name}/layer_{i}/attn/mhsa')(y, y)
        if vision_cfg.ls_init_value is not None:
            y = LayerScale(name=f'{name}/layer_{i}/attn/scale')(y)
        x = layers.add([x, y], name=f'{name}/layer_{i}/attn/add')

        y = layers.LayerNormalization(epsilon=ln_epsilon, name=f'{name}/layer_{i}/mlp/norm')(x)
        y = layers.Dense(int(vision_cfg.width * vision_cfg.mlp_ratio), name=f'{name}/layer_{i}/mlp/expand')(y)
        y = layers.Activation(mlp_act, name=f'{name}/layer_{i}/mlp/act')(y)
        y = layers.Dense(vision_cfg.width, name=f'{name}/layer_{i}/mlp/squeeze')(y)
        if vision_cfg.ls_init_value is not None:
            y = LayerScale(name=f'{name}/layer_{i}/mlp/scale')(y)
        x = layers.add([x, y], name=f'{name}/layer_{i}/mlp/add')

    x = layers.Activation('linear', name=f'{name}/head/in')(x)

    if vision_cfg.attentional_pool:
        x = AttentionalPooler(
            embed_dim, vision_cfg.attn_pooler_heads, vision_cfg.attn_pooler_queries, epsilon=ln_epsilon,
            name=f'{name}/head/attnpool')(x)
        x = layers.LayerNormalization(epsilon=ln_epsilon, name=f'{name}/head/norm')(x)
        pooled = GlobalPool(vision_cfg.pool_type, name=f'{name}/head/pool')(x)
    elif vision_cfg.ma_pool:
        pooled = layers.LayerNormalization(epsilon=ln_epsilon, name=f'{name}/head/norm')(x)
        pooled = MultiheadAttentionPooling(num_heads, name=f'{name}/head/mapool')(pooled)
        y = layers.LayerNormalization(epsilon=ln_epsilon, name=f'{name}/head/mapool/mlp/norm')(pooled)
        y = layers.Dense(int(vision_cfg.width * vision_cfg.mlp_ratio), name=f'{name}/head/mapool/mlp/expand')(y)
        y = layers.Activation(mlp_act, name=f'{name}/head/mapool/mlp/act')(y)
        y = layers.Dense(vision_cfg.width, name=f'{name}/head/mapool/mlp/squeeze')(y)
        pooled = layers.add([pooled, y], name=f'{name}/head/mapool/mlp/add')
        pooled = GlobalPool('first', name=f'{name}/head/pool')(pooled)
    elif vision_cfg.final_ln_after_pool:
        pooled = GlobalPool(vision_cfg.pool_type, name=f'{name}/head/pool')(x)
        pooled = layers.LayerNormalization(epsilon=ln_epsilon, name=f'{name}/head/norm')(pooled)
    else:
        x = layers.LayerNormalization(epsilon=ln_epsilon, name=f'{name}/head/norm')(x)
        pooled = GlobalPool(vision_cfg.pool_type, name=f'{name}/head/pool')(x)

    if not vision_cfg.ma_pool:
        pooled = layers.Dense(embed_dim, use_bias=False, name=f'{name}/head/proj')(pooled)

    pooled = layers.Activation('linear', name=f'{name}/head/out')(pooled)

    model = models.Model(inputs=image, outputs=pooled, name=name)

    return model


def GlobalPool(pool_type, name=None):
    if name is None:
        counter = backend.get_uid('global_pool')
        name = f'global_pool_{counter}'

    def apply(inputs):
        if 'avg' == pool_type:
            _, tokens = SplitClassToken(name=f'{name}/split')(inputs)
            pooled = layers.GlobalAvgPool1D(name=f'{name}/avg')(tokens)
        elif 'tok' == pool_type:
            pooled, _ = SplitClassToken(name=f'{name}/split')(inputs)
        elif 'first' == pool_type:
            pooled, _ = SplitClassToken(name=f'{name}/split')(inputs)
        else:
            pooled = inputs

        return pooled

    return apply



import numpy as np
from dataclasses import dataclass
from keras import layers, models
from typing import Optional, Tuple, Union
from tfclip.itsim import ImageTextSimilarity
from tfclip.text import TextTransformer
from tfclip.vision import VisionTransformer


@dataclass
class CLIPVisionCfg:
    layers: Union[Tuple[int, int, int, int], int] = 12
    width: int = 768
    head_width: int = 64
    mlp_ratio: float = 4.0
    patch_size: int = 16
    patch_bias: bool = False
    image_size: Union[Tuple[int, int], int] = 224

    embed_cls: bool = True
    ls_init_value: Optional[float] = None
    patch_dropout: float = 0.
    attentional_pool: bool = False
    attn_pooler_queries: int = 256
    attn_pooler_heads: int = 8
    ma_pool: bool = False
    no_ln_pre: bool = False
    pos_embed_type: str = 'learnable'
    final_ln_after_pool: bool = False
    pool_type: str = 'tok'
    output_tokens: bool = False
    act_kwargs: Optional[dict] = None
    norm_kwargs: Optional[dict] = None

    timm_model_name: Optional[str] = None
    timm_model_pretrained: bool = False
    timm_pool: str = 'avg'
    timm_proj: str = 'linear'
    timm_proj_bias: bool = False
    timm_drop: float = 0.
    timm_drop_path: Optional[float] = None


@dataclass
class CLIPTextCfg:
    context_length: int = 77
    vocab_size: int = 49408
    hf_tokenizer_name: Optional[str] = None
    tokenizer_kwargs: Optional[dict] = None

    width: int = 512
    heads: int = 8
    layers: int = 12
    mlp_ratio: float = 4.0
    ls_init_value: Optional[float] = None  # layer scale initial value
    embed_cls: bool = False
    pad_id: int = 0
    no_causal_mask: bool = False  # disable causal masking
    final_ln_after_pool: bool = False  # apply final LayerNorm after pooling
    pool_type: str = 'argmax'
    proj_bias: bool = False
    output_tokens: bool = False
    act_kwargs: dict = None
    norm_kwargs: dict = None

    # HuggingFace specific text tower config
    hf_model_name: Optional[str] = None
    hf_model_pretrained: bool = True
    hf_proj_type: str = 'mlp'
    hf_pooler_type: str = 'mean_pooler'  # attentional pooling for HF models

    # TensorFlow specific text tower config
    sp_tokenizer_name: Optional[str] = None
    hub_tokenizer_name: Optional[str] = None


def _build_vision_tower(embed_dim, vision_cfg, quick_gelu, img_mean, img_std):
    if isinstance(vision_cfg, dict):
        vision_cfg = CLIPVisionCfg(**vision_cfg)

    if vision_cfg.timm_model_name:
        raise ValueError(f'Unsupported image encoder: {vision_cfg.timm_model_name}')
    elif isinstance(vision_cfg.layers, (tuple, list)):
        raise ValueError(f'Unsupported image encoder: modified_resnet')
    else:
        model = VisionTransformer(embed_dim, vision_cfg, quick_gelu, img_mean, img_std)

    return model


def _build_text_tower(embed_dim, text_cfg, quick_gelu, custom=False):
    if isinstance(text_cfg, dict):
        text_cfg = CLIPTextCfg(**text_cfg)

    if text_cfg.hf_model_name:
        raise ValueError(f'Unsupported text encoder: {text_cfg.hf_model_name}')
    else:
        model = TextTransformer(embed_dim, text_cfg, quick_gelu, custom=custom)

    return model


def CLIP(
        embed_dim, vision_cfg, text_cfg, img_mean, img_std, quick_gelu=False, scale_init=-np.log(0.07),
        bias_init=None, custom_text=False):

    vision = _build_vision_tower(embed_dim, vision_cfg, quick_gelu, img_mean, img_std)
    text = _build_text_tower(embed_dim, text_cfg, quick_gelu, custom=custom_text)

    s = ImageTextSimilarity(scale_init, bias_init, name='head/sim')([vision.outputs[0], text.outputs[0]])
    s = layers.Activation('softmax', name='head/prob', dtype='float32')(s)

    model = models.Model(inputs=[vision.inputs[0], text.inputs[0]], outputs=s, name='clip')

    return model



import tensorflow as tf
from keras import layers
from keras.saving import register_keras_serializable
from keras.src.utils.tf_utils import shape_type_conversion


@register_keras_serializable(package='TFCLIP')
class AttentionalPooler(layers.Layer):
    def __init__(self, units, heads, queries, epsilon=1e-3, **kwargs):
        super().__init__(**kwargs)
        self.input_spec = layers.InputSpec(ndim=3)
        self.units = units
        self.heads = heads
        self.queries = queries
        self.epsilon = epsilon

    @shape_type_conversion
    def build(self, input_shape):
        channels = input_shape[-1]
        if channels is None:
            raise ValueError('Channel dimensions of the inputs should be defined. Found `None`.')
        self.input_spec = layers.InputSpec(ndim=3, axes={-1: channels})

        # noinspection PyAttributeOutsideInit
        self.query = self.add_weight(name='query', shape=(1, self.queries, self.units))

        # noinspection PyAttributeOutsideInit
        self.mhsa = layers.MultiHeadAttention(self.heads, self.units // self.heads, name='mhsa')

        # noinspection PyAttributeOutsideInit
        self.ln_q = layers.LayerNormalization(epsilon=self.epsilon, name='ln_q')

        # noinspection PyAttributeOutsideInit
        self.ln_k = layers.LayerNormalization(epsilon=self.epsilon, name='ln_k')

        super().build(input_shape)

    def call(self, inputs, *args, **kwargs):
        batch = tf.shape(inputs)[0]

        q = self.ln_q(self.query)
        q = tf.repeat(q, batch, axis=0)

        k = self.ln_k(inputs)
        x = self.mhsa(q, k)

        return x

    @shape_type_conversion
    def compute_output_shape(self, input_shape):
        return input_shape[:1] + (self.queries, self.units)

    def get_config(self):
        config = super().get_config()
        config.update({
            'units': self.units,
            'heads': self.heads,
            'queries': self.queries,
            'epsilon': self.epsilon
        })

        return config



import tensorflow as tf
from keras import layers
from keras.saving import register_keras_serializable
from keras.src.utils.tf_utils import shape_type_conversion


@register_keras_serializable(package='TFCLIP')
class AddClassToken(layers.Layer):
    def __init__(self, first=True, **kwargs):
        super().__init__(**kwargs)
        self.input_spec = layers.InputSpec(ndim=3)
        self.first = first

    @shape_type_conversion
    def build(self, input_shape):
        channels = input_shape[-1]
        if channels is None:
            raise ValueError('Channel dimensions of the inputs should be defined. Found `None`.')
        self.input_spec = layers.InputSpec(ndim=3, axes={-1: channels})

        # noinspection PyAttributeOutsideInit
        self.token = self.add_weight(name='token', shape=(1, 1, channels), initializer='zeros')

        super().build(input_shape)

    def call(self, inputs, *args, **kwargs):
        batch_size = tf.shape(inputs)[0]
        cls_token = tf.repeat(self.token, batch_size, axis=0)

        outputs = [cls_token, inputs] if self.first else [inputs, cls_token]
        outputs = tf.concat(outputs, axis=1)

        return outputs

    def compute_mask(self, inputs, mask=None):
        if mask is None:
            return None

        token_mask = tf.ones_like(mask[:, :1], dtype='bool')

        mask = [token_mask, mask] if self.first else [mask, token_mask]
        mask = tf.concat(mask, axis=1)

        return mask

    @shape_type_conversion
    def compute_output_shape(self, input_shape):
        length = input_shape[1]
        if length is None:
            return input_shape

        return input_shape[:1] + (length + 1,) + input_shape[2:]

    def get_config(self):
        config = super().get_config()
        config.update({'first': self.first})

        return config


@register_keras_serializable(package='TFCLIP')
class SplitClassToken(layers.Layer):
    def __init__(self, first=True, **kwargs):
        super().__init__(**kwargs)
        self.input_spec = layers.InputSpec(ndim=3)
        self.first = first

    @shape_type_conversion
    def build(self, input_shape):
        # noinspection PyAttributeOutsideInit
        self.channels = input_shape[-1]
        if self.channels is None:
            raise ValueError('Channel dimensions of the inputs should be defined. Found `None`.')
        self.input_spec = layers.InputSpec(ndim=3, axes={-1: self.channels})

        super().build(input_shape)

    def call(self, inputs, *args, **kwargs):
        if self.first:
            token, features = tf.split(inputs, [1, -1], axis=1)
        else:
            features, token = tf.split(inputs, [-1, 1], axis=1)
        token = tf.reshape(token, [-1, self.channels])

        return token, features

    @shape_type_conversion
    def compute_output_shape(self, input_shape):
        token_shape = input_shape[:1] + (self.channels,)
        features_shape = input_shape[:1] + (None if input_shape[1] is None else input_shape[1] - 1, self.channels)

        return token_shape, features_shape

    def get_config(self):
        config = super().get_config()
        config.update({'first': self.first})

        return config



import tensorflow as tf
from keras import layers
from keras.saving import register_keras_serializable
from keras.src.utils.tf_utils import shape_type_conversion


@register_keras_serializable(package='TFCLIP')
class MultiheadAttentionPooling(layers.Layer):
    def __init__(self, heads, **kwargs):
        super().__init__(**kwargs)
        self.input_spec = layers.InputSpec(ndim=3)
        self.heads = heads

    @shape_type_conversion
    def build(self, input_shape):
        channels = input_shape[-1]
        if channels is None:
            raise ValueError('Channel dimensions of the inputs should be defined. Found `None`.')
        self.input_spec = layers.InputSpec(ndim=3, axes={-1: channels})

        # noinspection PyAttributeOutsideInit
        self.query = self.add_weight(name='probe', shape=(1, 1, channels))

        # noinspection PyAttributeOutsideInit
        self.mhsa = layers.MultiHeadAttention(self.heads, channels // self.heads, name='mhsa')

        super().build(input_shape)

    def call(self, inputs, *args, **kwargs):
        batch = tf.shape(inputs)[0]

        q = tf.repeat(self.query, batch, axis=0)
        x = self.mhsa(q, inputs)

        return x

    @shape_type_conversion
    def compute_output_shape(self, input_shape):
        return input_shape[:1] + (1,) + input_shape[-1:]

    def get_config(self):
        config = super().get_config()
        config.update({'heads': self.heads})

        return config



import tensorflow as tf
from dataclasses import dataclass, asdict
from typing import Tuple, Union
from tfclip.constants import OPENAI_DATASET_MEAN, OPENAI_DATASET_STD
from tfclip.utils import to_2tuple


@dataclass
class PreprocessCfg:
    size: Union[int, Tuple[int, int]] = 224
    mode: str = 'RGB'
    mean: Tuple[float, ...] = OPENAI_DATASET_MEAN
    std: Tuple[float, ...] = OPENAI_DATASET_STD
    interpolation: str = 'bicubic'
    resize_mode: str = 'shortest'
    fill_color: int = 0

    def __post_init__(self):
        if 'RGB' != self.mode:
            raise ValueError(f'Unsupported image mode: {self.mode}')

    @property
    def num_channels(self):
        return 3

    @property
    def input_size(self):
        return (self.num_channels,) + to_2tuple(self.size)


_PREPROCESS_KEYS = set(asdict(PreprocessCfg()).keys())


def merge_preprocess_dict(base, overlay):
    if isinstance(base, PreprocessCfg):
        base_clean = asdict(base)
    else:
        base_clean = {k: v for k, v in base.items() if k in _PREPROCESS_KEYS}
    if overlay:
        overlay_clean = {k: v for k, v in overlay.items() if k in _PREPROCESS_KEYS and v is not None}
        base_clean.update(overlay_clean)
    return base_clean


def image_transform(image_size, interpolation_mode, resize_mode, name=None):
    if isinstance(image_size, int):
        image_size = (image_size, image_size)

    def apply(image):
        with tf.name_scope(name or 'image_transform'):
            image = tf.convert_to_tensor(image)
            if 3 != image.shape.rank or 3 != image.shape[-1]:
                raise ValueError(f'Expecting a single RGB image, got shape: {image.shape}')

            if 'bicubic' == interpolation_mode:
                interpolation = tf.image.ResizeMethod.BICUBIC
            elif 'bilinear' == interpolation_mode:
                interpolation = tf.image.ResizeMethod.BILINEAR
            else:
                raise ValueError(f'Unsupported interpolation mode: {interpolation_mode}')

            if 'squash' == resize_mode:
                image = tf.image.resize(image, image_size, method=interpolation)
            elif 'shortest' == resize_mode:
                source_size = tf.cast(tf.shape(image)[:2], 'float32')

                aspect_ratio = source_size / tf.cast(image_size, 'float32')
                aspect_ratio = tf.reduce_min(aspect_ratio)

                target_size = tf.cast(tf.round(source_size / aspect_ratio), 'int32')

                image = tf.image.resize(image, target_size, method=interpolation, antialias=True)
                image = tf.image.resize_with_crop_or_pad(image, image_size[0], image_size[1])
            else:
                raise ValueError(f'Unsupported resize mode: {resize_mode}')

            image = tf.clip_by_value(image, 0, 255)
            image = tf.round(image)
            image = tf.cast(image, 'uint8')

            return image

    return apply



import tensorflow as tf
from keras import layers
from keras.saving import register_keras_serializable
from keras.src.utils.tf_utils import shape_type_conversion


@register_keras_serializable(package='TFCLIP')
class TokenLastCausalMask(layers.Layer):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.input_spec = layers.InputSpec(ndim=2, dtype='int64')

    def call(self, inputs, *args, **kwargs):
        length = tf.shape(inputs)[1]
        length1 = length + 1

        mask = tf.linalg.band_part(tf.ones((1, length1, length1), tf.bool), -1, 0)

        cls_mask = (inputs > 0)[:, None]
        cls_mask = tf.pad(cls_mask, [[0, 0], [length, 0], [1, 0]], constant_values=True)

        return mask & cls_mask

    @shape_type_conversion
    def compute_output_shape(self, input_shape):
        if input_shape[1] is None:
            return input_shape[0], None, None

        return input_shape[0], input_shape[1] + 1, input_shape[1] + 1



from keras import layers, models
from tfclip.abspos import TextPositionEmbedding
from tfclip.clstok import AddClassToken
from tfclip.lscale import LayerScale
from tfclip.qgelu import q_gelu
from tfclip.tlcm import TokenLastCausalMask
from tfclip.txtpool import TextGlobalPool


def TextTransformer(embed_dim, text_cfg, quick_gelu, custom=False, name='text'):
    head_width = text_cfg.width // text_cfg.heads

    ln_epsilon = 1.001e-5
    if text_cfg.norm_kwargs:
        norm_kwargs_ = set(text_cfg.norm_kwargs.keys()) - {'eps'}
        if set(text_cfg.norm_kwargs.keys()) - {'eps'}:
            raise ValueError(f'Unsupported normalization arguments in config: {norm_kwargs_}')

        ln_epsilon = text_cfg.norm_kwargs.get('eps', ln_epsilon)

    if text_cfg.act_kwargs is not None:
        act_kwargs_ = set(text_cfg.act_kwargs.keys())
        raise ValueError(f'Unsupported activation arguments in config: {act_kwargs_}')

    use_cls = custom and text_cfg.embed_cls
    mlp_act = q_gelu if quick_gelu else 'gelu'

    # Define model inputs
    text = layers.Input(shape=(None,), dtype='int64', name=f'{name}/texts')

    # Define model pipeline
    x = text
    x = layers.Embedding(text_cfg.vocab_size, text_cfg.width, name=f'{name}/token/embed')(x)

    if use_cls:
        x = AddClassToken(first=False, name=f'{name}/token/cls')(x)
    x = TextPositionEmbedding(text_cfg.context_length + int(use_cls), name=f'{name}/token/pos')(x)

    if use_cls and not text_cfg.no_causal_mask:
        tlcm = TokenLastCausalMask(name=f'{name}/token/tlcm')(text)

    for i in range(text_cfg.layers):
        y = layers.LayerNormalization(epsilon=ln_epsilon, name=f'{name}/layer_{i}/attn/norm')(x)
        if use_cls:
            y = layers.MultiHeadAttention(text_cfg.heads, head_width, name=f'{name}/layer_{i}/attn/mhsa')(
                y, y, attention_mask=tlcm)
        else:
            y = layers.MultiHeadAttention(text_cfg.heads, head_width, name=f'{name}/layer_{i}/attn/mhsa')(
                y, y, use_causal_mask=not text_cfg.no_causal_mask)
        if text_cfg.ls_init_value is not None:
            y = LayerScale(name=f'{name}/layer_{i}/attn/scale')(y)
        x = layers.add([x, y], name=f'{name}/layer_{i}/attn/add')

        y = layers.LayerNormalization(epsilon=ln_epsilon, name=f'{name}/layer_{i}/mlp/norm')(x)
        y = layers.Dense(int(text_cfg.width * text_cfg.mlp_ratio), name=f'{name}/layer_{i}/mlp/expand')(y)
        y = layers.Activation(mlp_act, name=f'{name}/layer_{i}/mlp/act')(y)
        y = layers.Dense(text_cfg.width, name=f'{name}/layer_{i}/mlp/squeeze')(y)
        if text_cfg.ls_init_value is not None:
            y = LayerScale(name=f'{name}/layer_{i}/mlp/scale')(y)
        x = layers.add([x, y], name=f'{name}/layer_{i}/mlp/add')

    x = layers.Activation('linear', name=f'{name}/head/in')(x)

    if use_cls:
        # presence of appended cls embed (CoCa) overrides pool_type, always take cls token
        pooled, _ = TextGlobalPool('last', name=f'{name}/head/pool')([x, text])
        pooled = layers.LayerNormalization(epsilon=ln_epsilon, name=f'{name}/head/norm')(pooled)
    else:
        x = layers.LayerNormalization(epsilon=ln_epsilon, name=f'{name}/head/norm')(x)
        pooled, _ = TextGlobalPool(text_cfg.pool_type, name=f'{name}/head/pool')([x, text])
    pooled = layers.Dense(embed_dim, use_bias=text_cfg.proj_bias, name=f'{name}/head/proj')(pooled)

    pooled = layers.Activation('linear', name=f'{name}/head/out')(pooled)

    model = models.Model(inputs=text, outputs=pooled, name=name)

    return model



from tfclip.constants import OPENAI_DATASET_MEAN, OPENAI_DATASET_STD, INCEPTION_MEAN, INCEPTION_STD, \
    IMAGENET_MEAN, IMAGENET_STD


def _pcfg(version='1.0.0', sha256=None, **kwargs):
    # OpenAI / OpenCLIP defaults
    return {
        'mean': OPENAI_DATASET_MEAN,
        'std': OPENAI_DATASET_STD,
        'interpolation': 'bicubic',
        'resize_mode': 'shortest',
        'version': version,
        'sha256': sha256,
        **kwargs,
    }


def _slpcfg(version='1.0.0', sha256=None, **kwargs):
    # SiGLIP defaults
    return {
        'mean': INCEPTION_MEAN,
        'std': INCEPTION_STD,
        'interpolation': 'bicubic',
        'resize_mode': 'squash',
        'version': version,
        'sha256': sha256,
        **kwargs,
    }


def _apcfg(version='1.0.0', sha256=None, **kwargs):
    # CLIPA defaults
    return {
        'mean': IMAGENET_MEAN,
        'std': IMAGENET_STD,
        'interpolation': 'bilinear',
        'resize_mode': 'squash',
        'version': version,
        'sha256': sha256,
        **kwargs,
    }


# _RN50 = dict(
#     openai=_pcfg(),
#     yfcc15m=_pcfg(),
#     cc12m=_pcfg(),
# )
#
# _RN50_quickgelu = dict(
#     openai=_pcfg(),
#     yfcc15m=_pcfg(),
#     cc12m=_pcfg(),
# )
#
# _RN101 = dict(
#     openai=_pcfg(),
#     yfcc15m=_pcfg(),
# )
#
# _RN101_quickgelu = dict(
#     openai=_pcfg(),
#     yfcc15m=_pcfg(),
# )
#
# _RN50x4 = dict(
#     openai=_pcfg(),
# )
#
# _RN50x16 = dict(
#     openai=_pcfg(),
# )
#
# _RN50x64 = dict(
#     openai=_pcfg(),
# )

_VITB32 = dict(
    # openai=_pcfg(),  # OpenAI models were trained with QuickGELU
    laion400m_e31=_pcfg(sha256='320662fda288042c60a3b6175fb4c31f7b71803d9ce6d6ceb1fd69f5c9340b76'),
    laion400m_e32=_pcfg(sha256='868b1597cc633a9ed2a0aecb07fc77b1a939b6ef11a719bfdacbab19f524fe12 '),
    laion2b_e16=_pcfg(sha256='649b1d8ed57828bebd1256b7bd81e51956cb2a9fa6f882fa20abc216811b8917'),
    laion2b_s34b_b79k=_pcfg(sha256='c9b3aa9965d2dc9d34fe9751ae3b8b00b56cf30bcf260ffab2cf9c7cb6ecc38a'),
    # DataComp-XL models
    datacomp_xl_s13b_b90k=_pcfg(sha256='d550a54e17fc9a9af9bbe620ce48ae7c20f8d38761f6e5467a2be1b7688fe60f'),
    # DataComp-M models
    datacomp_m_s128m_b4k=_pcfg(sha256='b081e03fa06c4dec5ae383887b2072ad22197ce47a74b5d668a136c313aa91d1'),
    commonpool_m_clip_s128m_b4k=_pcfg(sha256='807883587d8a54e5aad62268f9736e288ed8dabb9fdaeb1562382d9623c3fea7'),
    commonpool_m_laion_s128m_b4k=_pcfg(sha256='7062564cb76671d629018f203114664a34039d2153cfc9dcf1150023b2acd6b5'),
    commonpool_m_image_s128m_b4k=_pcfg(sha256='0b2baccc96cd6717bf945bd6004bc552cee528461e861e8a53b9f1d3ad40f99a'),
    commonpool_m_text_s128m_b4k=_pcfg(sha256='ae9d94b8ff4083be9798840d474478fb2102f2f93e6125710684e46543261057'),
    commonpool_m_basic_s128m_b4k=_pcfg(sha256='b105cef0763b8aea4fd73b6526f2392180a41c485646f0704fd3386092905e75'),
    commonpool_m_s128m_b4k=_pcfg(sha256='c6cd5fcac6664ea520baa38337f769ba3f8516cd551f0b90511cc83be63415e2'),
    # DataComp-S models
    datacomp_s_s13m_b4k=_pcfg(sha256='cdc5303490ed6df01475ac0ccedb3435ae9ead16d22edcc25c65492e573d5a98'),
    commonpool_s_clip_s13m_b4k=_pcfg(sha256='35e3073a991db2cf92129e71da6aae98da560f268d867199d5ed7cc07abeac53'),
    commonpool_s_laion_s13m_b4k=_pcfg(sha256='aca756a042efffdd9f2f1ae64cfb51dcb8cff1ccec15bae1bbcb2db735e85bf5'),
    commonpool_s_image_s13m_b4k=_pcfg(sha256='cdc5303490ed6df01475ac0ccedb3435ae9ead16d22edcc25c65492e573d5a98'),
    commonpool_s_text_s13m_b4k=_pcfg(sha256='ecb478df74139d50fea4602dc7f2852b6f0722d83583ff22ab537a3e80939a79'),
    commonpool_s_basic_s13m_b4k=_pcfg(sha256='d74f3485d2a59b32b8a8bc44524d62ea0e01596cdd8d2240f01934072cc983de'),
    commonpool_s_s13m_b4k=_pcfg(sha256='d4333ca7b35e67444f18c57e35f5c2de8e17ee76f859fc1efd5bfa13fa0ad776'),
)

_VITB32_quickgelu = dict(
    openai=_pcfg(sha256='2e0f33c7e468ddaf7e5f615ad557faaaca16ba1f1cf49558953f8bda6cd16006'),
    laion400m_e31=_pcfg(sha256='320662fda288042c60a3b6175fb4c31f7b71803d9ce6d6ceb1fd69f5c9340b76'),
    laion400m_e32=_pcfg(sha256='868b1597cc633a9ed2a0aecb07fc77b1a939b6ef11a719bfdacbab19f524fe12'),
    metaclip_400m=_pcfg(sha256='57a88f153fbae708ef901ecf2ad8288dad702f5693793515f0371fdcdf480e50'),
    metaclip_fullcc=_pcfg(sha256='cce1320dacdf0b44d2c768e62518b7e37281ff264cfea64f986a6eeb586fe955'),
)

_VITB32_256 = dict(
    datacomp_s34b_b86k=_pcfg(sha256='24aba612bd1d6b50908be3f607c685a1b6a2b962f2d38dfec5dec69fd5cd1d83'),
)

_VITB16 = dict(
    # openai=_pcfg(),  # OpenAI models were trained with QuickGELU
    laion400m_e31=_pcfg(sha256='14796113406aaf645350d0258b937a2e5729d5c851b6b9bed4c5d4b8aefe124f'),
    laion400m_e32=_pcfg(sha256='74799b8f05c38a6364ad999afb459b9fafa073e35fbbb063a6a79ab4d8f30ac9'),
    laion2b_s34b_b88k=_pcfg(sha256='fe66282f80d7243afb7a463db56cf0d3f1bcce0abf665b6af5134544be451700'),
    # DataComp-XL models
    datacomp_xl_s13b_b90k=_pcfg(sha256='20d5f78b7ea7aad0f9727d7c0c1e27be06a16ecbb2130ada290e7a166b57ad23'),
    # DataComp-L models
    datacomp_l_s1b_b8k=_pcfg(sha256='3ef342da977ec12d8f2d25d2f3af9e0f98ecaaa0069aac255369bd09133dc8d8'),
    commonpool_l_clip_s1b_b8k=_pcfg(sha256='6a42ac82a9dda60ec082855cd746a90eb2dbc566fc5e47539cd6bcf4b4d5159b'),
    commonpool_l_laion_s1b_b8k=_pcfg(sha256='14f09c24bc8a7516d9147b34cb0bd7f058c43339e7c3b10cd10e55a1cc82b995'),
    commonpool_l_image_s1b_b8k=_pcfg(sha256='12679e00d2be224de4eb971534b0e64943b151b8570387bd390c51b6dae3a45b'),
    commonpool_l_text_s1b_b8k=_pcfg(sha256='bb91fe106f4eed27dc672ad12ba59b27f43cff0651a27bceeffd7230a4307cff'),
    commonpool_l_basic_s1b_b8k=_pcfg(sha256='7f6df7729ebacd07d8a27bd69b5c99755731328ee81f5f53d5475fb120c67396'),
    commonpool_l_s1b_b8k=_pcfg(sha256='73f2da6dc01575c3ddb17ed1b459b1ef07dc9fe5dc54e68d64bcdd1c8407559a'),
)

_VITB16_quickgelu = dict(
    # OpenAI models were trained with QuickGELU
    openai=_pcfg(sha256='a2f7a56319a2be0e8f226480ccf2d29ceb20e5e414bbd158a883664db9276cac'),
    metaclip_400m=_pcfg(sha256='2c186d9d1e2e69c7caf6bf3e2c076df9032ff53439a5b8df5411f829863b9963'),
    metaclip_fullcc=_pcfg(sha256='4575651e9fa5259fe12e34b747795fb7122e1fa3051eb51e40e25ca29aadba16'),
)

_VITB16_PLUS_240 = dict(
    laion400m_e31=_pcfg(sha256='6817324ef268d8cd790cd778b4f1f12458323cd7bba0edad4f9510582d0367f3'),
    laion400m_e32=_pcfg(sha256='c2f976d855c403cf4bce9d8e0dde9cff965d97e64864a061ae2b3bae6875594e'),
)

_VITL14 = dict(
    # openai=_pcfg(),  # OpenAI models were trained with QuickGELU
    laion400m_e31=_pcfg(sha256='42f8c8fc92bd60272eb6d3784df5fdfed8b3ba73f7d9d5c7be635a3465bd7150'),
    laion400m_e32=_pcfg(sha256='bc883c6f3bd007e6c0ee4c16bea68220b4f0d14deb3ba3c246ad8d954551e467'),
    laion2b_s32b_b82k=_pcfg(
        mean=INCEPTION_MEAN, std=INCEPTION_STD,
        sha256='6e882e18abf7bd6111aa3651dcbc265cb99b784ce40bb7b7e73ae0ed39586614'),
    # DataComp-XL models
    datacomp_xl_s13b_b90k=_pcfg(sha256='a18b7ae3a2ef324fb5115b0f742ae15f49e1bd8318eb54371a2c45a1b681a1c0'),
    commonpool_xl_clip_s13b_b90k=_pcfg(sha256='1e1987048d3d2e008898f81c8ff54f04b9c12cf23dc9ce9006897d31ee53f157'),
    commonpool_xl_laion_s13b_b90k=_pcfg(sha256='1193dbce1dfcaef9f9005e2d931f22c9d677af02c18c89f973ab38b5aa82f3a4'),
    commonpool_xl_s13b_b90k=_pcfg(sha256='b7515ce1a933ad6d8c35ff0a8a7590f76e7a02a18bb71697da101ba1f36d4b39'),
)

_VITL14_quickgelu = dict(
    # OpenAI models were trained with QuickGELU
    openai=_pcfg(sha256='cd11e3ebc37d88a89e9fb788fe32472791de23813666b1531ab971d3aae327c5'),
    metaclip_400m=_pcfg(sha256='d992eb4050ed92f50fc92889e8a3417d1a35d338676930931544617c81244ec5'),
    metaclip_fullcc=_pcfg(sha256='383266d4e77954cd8602a92fb095103ed4d0118b51bb17d739d12a336cc7c149'),
)

_VITL14_336_quickgelu = dict(
    # OpenAI models were trained with QuickGELU
    openai=_pcfg(sha256='ad628eb668585f44e59fc151da40f2fa47c4d4e32336259f8fc1522ee40088d9'),
)

_VITH14 = dict(
    laion2b_s32b_b79k=_pcfg(),
)

_VITH14_quickgelu = dict(
    metaclip_fullcc=_pcfg(),
)

_VITg14 = dict(
    laion2b_s12b_b42k=_pcfg(),
    laion2b_s34b_b88k=_pcfg(),
)

_VITbigG14 = dict(
    laion2b_s39b_b160k=_pcfg(),
)

# _robertaViTB32 = dict(
#     laion2b_s12b_b32k=_pcfg(),
# )
#
# _xlmRobertaBaseViTB32 = dict(
#     laion5b_s13b_b90k=_pcfg(),
# )
#
# _xlmRobertaLargeFrozenViTH14 = dict(
#     frozen_laion5b_s13b_b90k=_pcfg(),
# )
#
# _convnext_base = dict(
#     laion400m_s13b_b51k=_pcfg(),
# )
#
# _convnext_base_w = dict(
#     laion2b_s13b_b82k=_pcfg(),
#     laion2b_s13b_b82k_augreg=_pcfg(),
#     laion_aesthetic_s13b_b82k=_pcfg(),
# )
#
# _convnext_base_w_320 = dict(
#     laion_aesthetic_s13b_b82k=_pcfg(),
#     laion_aesthetic_s13b_b82k_augreg=_pcfg(),
# )
#
# _convnext_large_d = dict(
#     laion2b_s26b_b102k_augreg=_pcfg(),
# )
#
# _convnext_large_d_320 = dict(
#     laion2b_s29b_b131k_ft=_pcfg(),
#     laion2b_s29b_b131k_ft_soup=_pcfg(),
# )
#
# _convnext_xxlarge = dict(
#     laion2b_s34b_b82k_augreg=_pcfg(),
#     laion2b_s34b_b82k_augreg_rewind=_pcfg(),
#     laion2b_s34b_b82k_augreg_soup=_pcfg(),
# )

_coca_VITB32 = dict(
    laion2b_s13b_b90k=_pcfg(sha256='ff9b00049daad95ffc724ae8b57f5760ea99e6f41b7e4ad8b69198198dd6f767'),
    mscoco_finetuned_laion2b_s13b_b90k=_pcfg(sha256='e8d09d9feb7b6945ad241655639729bf72f54efeda1f153b05b71b508be435fa')
)

_coca_VITL14 = dict(
    laion2b_s13b_b90k=_pcfg(sha256='1a42f67fb75668e55c2a507022ea490de7ee946786afbe37cb6b992385440476'),
    mscoco_finetuned_laion2b_s13b_b90k=_pcfg(sha256='fc85f824dd2a70be272652b7fa83327e96b91675a8431cfe843d13ab69043411')
)

_PRETRAINED = {
    # 'RN50': _RN50,
    # 'RN50-quickgelu': _RN50_quickgelu,
    # 'RN101': _RN101,
    # 'RN101-quickgelu': _RN101_quickgelu,
    # 'RN50x4': _RN50x4,
    # 'RN50x16': _RN50x16,
    # 'RN50x64': _RN50x64,

    'ViT-B-32': _VITB32,
    'ViT-B-32-256': _VITB32_256,
    'ViT-B-32-quickgelu': _VITB32_quickgelu,
    'ViT-B-16': _VITB16,
    'ViT-B-16-quickgelu': _VITB16_quickgelu,
    'ViT-B-16-plus-240': _VITB16_PLUS_240,
    'ViT-L-14': _VITL14,
    'ViT-L-14-quickgelu': _VITL14_quickgelu,
    'ViT-L-14-336-quickgelu': _VITL14_336_quickgelu,  # OpenAI models were trained with QuickGELU
    'ViT-H-14': _VITH14,
    'ViT-H-14-quickgelu': _VITH14_quickgelu,
    'ViT-g-14': _VITg14,
    'ViT-bigG-14': _VITbigG14,

    # 'roberta-ViT-B-32': _robertaViTB32,
    # 'xlm-roberta-base-ViT-B-32': _xlmRobertaBaseViTB32,
    # 'xlm-roberta-large-ViT-H-14': _xlmRobertaLargeFrozenViTH14,

    # 'convnext_base': _convnext_base,
    # 'convnext_base_w': _convnext_base_w,
    # 'convnext_base_w_320': _convnext_base_w_320,
    # 'convnext_large_d': _convnext_large_d,
    # 'convnext_large_d_320': _convnext_large_d_320,
    # 'convnext_xxlarge': _convnext_xxlarge,

    'coca_ViT-B-32': _coca_VITB32,
    'coca_ViT-L-14': _coca_VITL14,

    # 'EVA01-g-14': dict(
    #     # from QuanSun/EVA-CLIP/EVA01_CLIP_g_14_psz14_s11B.pt
    #     laion400m_s11b_b41k=_pcfg(),
    # ),
    # 'EVA01-g-14-plus': dict(
    #     # from QuanSun/EVA-CLIP/EVA01_CLIP_g_14_plus_psz14_s11B.pt
    #     merged2b_s11b_b114k=_pcfg(),
    # ),
    # 'EVA02-B-16': dict(
    #     # from QuanSun/EVA-CLIP/EVA02_CLIP_B_psz16_s8B.pt
    #     merged2b_s8b_b131k=_pcfg(),
    # ),
    # 'EVA02-L-14': dict(
    #     # from QuanSun/EVA-CLIP/EVA02_CLIP_L_psz14_s4B.pt
    #     merged2b_s4b_b131k=_pcfg(),
    # ),
    # 'EVA02-L-14-336': dict(
    #     # from QuanSun/EVA-CLIP/EVA02_CLIP_L_336_psz14_s6B.pt
    #     merged2b_s6b_b61k=_pcfg(),
    # ),
    # 'EVA02-E-14': dict(
    #     # from QuanSun/EVA-CLIP/EVA02_CLIP_E_psz14_s4B.pt
    #     laion2b_s4b_b115k=_pcfg(),
    # ),
    # 'EVA02-E-14-plus': dict(
    #     # from QuanSun/EVA-CLIP/EVA02_CLIP_E_psz14_plus_s9B.pt
    #     laion2b_s9b_b144k=_pcfg(),
    # ),

    'ViT-B-16-SigLIP': dict(
        webli=_slpcfg(sha256='cb71518557f6e047e24d92aeb72e78b519b151dadaed5123d28cb71940cf4c66'),
    ),
    'ViT-B-16-SigLIP-256': dict(
        webli=_slpcfg(sha256='850a9b318b893070d4a0b35ddb237b50281fc155d84b234e463c1d3714e88505'),
    ),
    'ViT-B-16-SigLIP-i18n-256': dict(
        webli=_slpcfg(sha256='9bd33305441b7603543d13142752c0407cecda677a5f94bd025f0754a20eb50b'),
    ),
    'ViT-B-16-SigLIP-384': dict(
        webli=_slpcfg(sha256='9637f1ed2ccfbbccf54bc0638c366c7e42639c0374b5d90930dfd7c3f3a159d7'),
    ),
    'ViT-B-16-SigLIP-512': dict(
        webli=_slpcfg(sha256='ab13bcfc5c95080056a83adc37ecbdbf9877a5e3c4e4f20c203c7e7e10bd5e75'),
    ),
    'ViT-L-16-SigLIP-256': dict(
        webli=_slpcfg(),
    ),
    'ViT-L-16-SigLIP-384': dict(
        webli=_slpcfg(),
    ),
    'ViT-SO400M-14-SigLIP': dict(
        webli=_slpcfg(),
    ),
    'ViT-SO400M-14-SigLIP-384': dict(
        webli=_slpcfg(),
    ),

    'ViT-L-14-CLIPA': dict(
        datacomp1b=_apcfg(sha256='c6b750409035b20d8d43a260a590a643cd2bdce7e00a19074266228e799b36da'),
    ),
    'ViT-L-14-CLIPA-336': dict(
        datacomp1b=_apcfg(sha256='10c59fa86d38786927edf6b8a9c40ae65763ec76026e13f84c383e47f1ba722a'),
    ),
    'ViT-H-14-CLIPA': dict(
        datacomp1b=_apcfg(),
    ),
    'ViT-H-14-CLIPA-336-quickgelu': dict(
        laion2b=_apcfg(),
        datacomp1b=_apcfg(),
    ),
    'ViT-bigG-14-CLIPA': dict(
        datacomp1b=_apcfg(),
    ),
    'ViT-bigG-14-CLIPA-336': dict(
        datacomp1b=_apcfg(),
    ),

    # 'nllb-clip-base': dict(
    #     v1=_pcfg(),
    # ),
    # 'nllb-clip-large': dict(
    #     v1=_pcfg(),
    # )
}


def _clean_tag(tag):
    return tag.lower().replace('-', '_')


def list_pretrained(as_str=False):
    return [':'.join([k, t]) if as_str else (k, t) for k in _PRETRAINED.keys() for t in _PRETRAINED[k].keys()]


def list_pretrained_models_by_tag(tag):
    models = []
    tag = _clean_tag(tag)
    for k in _PRETRAINED.keys():
        if tag in _PRETRAINED[k]:
            models.append(k)

    return models


def list_pretrained_tags_by_model(model):
    tags = []
    if model in _PRETRAINED:
        tags.extend(_PRETRAINED[model].keys())
    return tags


def get_pretrained_cfg(model, tag):
    if model not in _PRETRAINED:
        return {}
    if tag is None:
        return {}

    model_pretrained = _PRETRAINED[model]

    return model_pretrained.get(_clean_tag(tag), {})


def get_pretrained_url(model, tag):
    tag = _clean_tag(tag)
    cfg = get_pretrained_cfg(model, tag)

    version = cfg.get('version', None)
    if version is None:
        return None

    sha256 = cfg.get('sha256', None)
    if sha256 is None:
        return None

    return f'https://github.com/shkarupa-alex/tfclip/releases/download/{version}/{model}__{tag}__{sha256}.h5'



import numpy as np
import tensorflow as tf
from keras import initializers, layers
from keras.saving import register_keras_serializable
from keras.src.utils.tf_utils import shape_type_conversion


@register_keras_serializable(package='TFCLIP')
class ImageTextSimilarity(layers.Layer):
    def __init__(self, scale_init, bias_init, **kwargs):
        super().__init__(**kwargs)
        self.input_spec = [layers.InputSpec(ndim=2), layers.InputSpec(ndim=2)]

        self.scale_init = scale_init
        self.bias_init = bias_init

    @shape_type_conversion
    def build(self, input_shape):
        # noinspection PyAttributeOutsideInit
        self.scale = self.add_weight(
            'scale', shape=[],
            initializer=initializers.constant(self.scale_init),
            constraint=lambda s: tf.minimum(s, np.log(100., dtype=self.dtype)))

        if self.bias_init:
            # noinspection PyAttributeOutsideInit
            self.bias = self.add_weight(
                'bias', shape=[],
                initializer=initializers.constant(self.bias_init))

        super().build(input_shape)

    def call(self, inputs, *args, **kwargs):
        image, text = inputs

        image /= tf.norm(image, axis=-1, keepdims=True)
        text /= tf.norm(text, axis=-1, keepdims=True)

        outputs = tf.matmul(image * tf.exp(self.scale), text, transpose_b=True)
        if self.bias_init:
            outputs += self.bias

        return outputs

    @shape_type_conversion
    def compute_output_shape(self, input_shape):
        return input_shape[0][0], input_shape[1][0]

    def get_config(self):
        config = super().get_config()
        config.update({
            'scale_init': self.scale_init,
            'bias_init': self.bias_init,
        })

        return config
